@@ -4,6 +4,7 @@ import itertools
 import logging
 import math
 import re
+import sys
 from collections import namedtuple
 from contextlib import suppress
 from functools import lru_cache, partial
@@ -12,7 +13,7 @@ from keyword import iskeyword
 import attr
 import pycodestyle
 
-__version__ = "22.7.1"
+__version__ = "23.1.14"
 
 LOG = logging.getLogger("flake8.bugbear")
 CONTEXTFUL_NODES = (
@@ -168,8 +169,10 @@ class BugBearChecker:
                 return True
 
         LOG.info(
-            "Optional warning %s not present in selected warnings: %r. Not "
-            "firing it at all.",
+            (
+                "Optional warning %s not present in selected warnings: %r. Not "
+                "firing it at all."
+            ),
             code,
             self.options.select,
         )
@@ -280,7 +283,7 @@ class BugBearVisitor(ast.NodeVisitor):
             names = [_to_name_str(e) for e in node.type.elts]
             as_ = " as " + node.name if node.name is not None else ""
             if len(names) == 0:
-                vs = ("`except (){}:`".format(as_),)
+                vs = (f"`except (){as_}:`",)
                 self.errors.append(B001(node.lineno, node.col_offset, vars=vs))
             elif len(names) == 1:
                 self.errors.append(B013(node.lineno, node.col_offset, vars=names))
@@ -354,6 +357,13 @@ class BugBearVisitor(ast.NodeVisitor):
                 ):
                     self.errors.append(B010(node.lineno, node.col_offset))
 
+            self.check_for_b026(node)
+
+        self.check_for_b905(node)
+        self.generic_visit(node)
+
+    def visit_Module(self, node):
+        self.check_for_b018(node)
         self.generic_visit(node)
 
     def visit_Assign(self, node):
@@ -410,16 +420,19 @@ class BugBearVisitor(ast.NodeVisitor):
         self.check_for_b018(node)
         self.check_for_b019(node)
         self.check_for_b021(node)
+        self.check_for_b906(node)
         self.generic_visit(node)
 
     def visit_ClassDef(self, node):
         self.check_for_b903(node)
         self.check_for_b018(node)
         self.check_for_b021(node)
+        self.check_for_b024_and_b027(node)
         self.generic_visit(node)
 
     def visit_Try(self, node):
         self.check_for_b012(node)
+        self.check_for_b025(node)
         self.generic_visit(node)
 
     def visit_Compare(self, node):
@@ -436,6 +449,10 @@ class BugBearVisitor(ast.NodeVisitor):
         self.check_for_b022(node)
         self.generic_visit(node)
 
+    def visit_JoinedStr(self, node):
+        self.check_for_b028(node)
+        self.generic_visit(node)
+
     def check_for_b005(self, node):
         if node.func.attr not in B005.methods:
             return  # method name doesn't match
@@ -504,6 +521,7 @@ class BugBearVisitor(ast.NodeVisitor):
 
     def check_for_b017(self, node):
         """Checks for use of the evil syntax 'with assertRaises(Exception):'
+        or 'with pytest.raises(Exception)'.
 
         This form of assertRaises will catch everything that subclasses
         Exception, which happens to be the vast majority of Python internal
@@ -513,10 +531,21 @@ class BugBearVisitor(ast.NodeVisitor):
         """
         item = node.items[0]
         item_context = item.context_expr
+
         if (
             hasattr(item_context, "func")
-            and hasattr(item_context.func, "attr")
-            and item_context.func.attr == "assertRaises"
+            and (
+                (
+                    hasattr(item_context.func, "attr")
+                    and item_context.func.attr == "assertRaises"
+                )
+                or (
+                    isinstance(item_context.func, ast.Attribute)
+                    and item_context.func.attr == "raises"
+                    and isinstance(item_context.func.value, ast.Name)
+                    and item_context.func.value.id == "pytest"
+                )
+            )
             and len(item_context.args) == 1
             and isinstance(item_context.args[0], ast.Name)
             and item_context.args[0].id == "Exception"
@@ -564,7 +593,7 @@ class BugBearVisitor(ast.NodeVisitor):
                 n = targets.names[name][0]
                 self.errors.append(B020(n.lineno, n.col_offset, vars=(name,)))
 
-    def check_for_b023(self, loop_node):
+    def check_for_b023(self, loop_node):  # noqa: C901
         """Check that functions (including lambdas) do not use loop variables.
 
         https://docs.python-guide.org/writing/gotchas/#late-binding-closures from
@@ -580,9 +609,41 @@ class BugBearVisitor(ast.NodeVisitor):
         # implement this "backwards": first we find all the candidate variable
         # uses, and then if there are any we check for assignment of those names
         # inside the loop body.
+        safe_functions = []
         suspicious_variables = []
         for node in ast.walk(loop_node):
-            if isinstance(node, FUNCTION_NODES):
+            # check if function is immediately consumed to avoid false alarm
+            if isinstance(node, ast.Call):
+                # check for filter&reduce
+                if (
+                    isinstance(node.func, ast.Name)
+                    and node.func.id in ("filter", "reduce", "map")
+                ) or (
+                    isinstance(node.func, ast.Attribute)
+                    and node.func.attr == "reduce"
+                    and isinstance(node.func.value, ast.Name)
+                    and node.func.value.id == "functools"
+                ):
+                    for arg in node.args:
+                        if isinstance(arg, FUNCTION_NODES):
+                            safe_functions.append(arg)
+
+                # check for key=
+                for keyword in node.keywords:
+                    if keyword.arg == "key" and isinstance(
+                        keyword.value, FUNCTION_NODES
+                    ):
+                        safe_functions.append(keyword.value)
+
+            # mark `return lambda: x` as safe
+            # does not (currently) check inner lambdas in a returned expression
+            # e.g. `return (lambda: x, )
+            if isinstance(node, ast.Return):
+                if isinstance(node.value, FUNCTION_NODES):
+                    safe_functions.append(node.value)
+
+            # find unsafe functions
+            if isinstance(node, FUNCTION_NODES) and node not in safe_functions:
                 argnames = {
                     arg.arg for arg in ast.walk(node.args) if isinstance(arg, ast.arg)
                 }
@@ -590,16 +651,19 @@ class BugBearVisitor(ast.NodeVisitor):
                     body_nodes = ast.walk(node.body)
                 else:
                     body_nodes = itertools.chain.from_iterable(map(ast.walk, node.body))
+                errors = []
                 for name in body_nodes:
-                    if (
-                        isinstance(name, ast.Name)
-                        and name.id not in argnames
-                        and isinstance(name.ctx, ast.Load)
-                    ):
-                        err = B023(name.lineno, name.col_offset, vars=(name.id,))
-                        if err not in self._b023_seen:
-                            self._b023_seen.add(err)  # dedupe across nested loops
-                            suspicious_variables.append(err)
+                    if isinstance(name, ast.Name) and name.id not in argnames:
+                        if isinstance(name.ctx, ast.Load):
+                            errors.append(
+                                B023(name.lineno, name.col_offset, vars=(name.id,))
+                            )
+                        elif isinstance(name.ctx, ast.Store):
+                            argnames.add(name.id)
+                for err in errors:
+                    if err.vars[0] not in argnames and err not in self._b023_seen:
+                        self._b023_seen.add(err)  # dedupe across nested loops
+                        suspicious_variables.append(err)
 
         if suspicious_variables:
             reassigned_in_loop = set(self._get_assigned_names(loop_node))
@@ -608,6 +672,104 @@ class BugBearVisitor(ast.NodeVisitor):
             if reassigned_in_loop.issuperset(err.vars):
                 self.errors.append(err)
 
+    def check_for_b024_and_b027(self, node: ast.ClassDef):  # noqa: C901
+        """Check for inheritance from abstract classes in abc and lack of
+        any methods decorated with abstract*"""
+
+        def is_abc_class(value, name="ABC"):
+            # class foo(metaclass = [abc.]ABCMeta)
+            if isinstance(value, ast.keyword):
+                return value.arg == "metaclass" and is_abc_class(value.value, "ABCMeta")
+            # class foo(ABC)
+            # class foo(abc.ABC)
+            return (isinstance(value, ast.Name) and value.id == name) or (
+                isinstance(value, ast.Attribute)
+                and value.attr == name
+                and isinstance(value.value, ast.Name)
+                and value.value.id == "abc"
+            )
+
+        def is_abstract_decorator(expr):
+            return (isinstance(expr, ast.Name) and expr.id[:8] == "abstract") or (
+                isinstance(expr, ast.Attribute) and expr.attr[:8] == "abstract"
+            )
+
+        def is_overload(expr):
+            return (isinstance(expr, ast.Name) and expr.id == "overload") or (
+                isinstance(expr, ast.Attribute) and expr.attr == "overload"
+            )
+
+        def empty_body(body) -> bool:
+            def is_str_or_ellipsis(node):
+                # ast.Ellipsis and ast.Str used in python<3.8
+                return isinstance(node, (ast.Ellipsis, ast.Str)) or (
+                    isinstance(node, ast.Constant)
+                    and (node.value is Ellipsis or isinstance(node.value, str))
+                )
+
+            # Function body consist solely of `pass`, `...`, and/or (doc)string literals
+            return all(
+                isinstance(stmt, ast.Pass)
+                or (isinstance(stmt, ast.Expr) and is_str_or_ellipsis(stmt.value))
+                for stmt in body
+            )
+
+        # don't check multiple inheritance
+        # https://github.com/PyCQA/flake8-bugbear/issues/277
+        if len(node.bases) + len(node.keywords) > 1:
+            return
+
+        # only check abstract classes
+        if not any(map(is_abc_class, (*node.bases, *node.keywords))):
+            return
+
+        has_abstract_method = False
+
+        for stmt in node.body:
+            # https://github.com/PyCQA/flake8-bugbear/issues/293
+            # Ignore abc's that declares a class attribute that must be set
+            if isinstance(stmt, (ast.AnnAssign, ast.Assign)):
+                has_abstract_method = True
+                continue
+
+            # only check function defs
+            if not isinstance(stmt, (ast.FunctionDef, ast.AsyncFunctionDef)):
+                continue
+
+            has_abstract_decorator = any(
+                map(is_abstract_decorator, stmt.decorator_list)
+            )
+
+            has_abstract_method |= has_abstract_decorator
+
+            if (
+                not has_abstract_decorator
+                and empty_body(stmt.body)
+                and not any(map(is_overload, stmt.decorator_list))
+            ):
+                self.errors.append(
+                    B027(stmt.lineno, stmt.col_offset, vars=(stmt.name,))
+                )
+
+        if not has_abstract_method:
+            self.errors.append(B024(node.lineno, node.col_offset, vars=(node.name,)))
+
+    def check_for_b026(self, call: ast.Call):
+        if not call.keywords:
+            return
+
+        starreds = [arg for arg in call.args if isinstance(arg, ast.Starred)]
+        if not starreds:
+            return
+
+        first_keyword = call.keywords[0].value
+        for starred in starreds:
+            if (starred.lineno, starred.col_offset) > (
+                first_keyword.lineno,
+                first_keyword.col_offset,
+            ):
+                self.errors.append(B026(starred.lineno, starred.col_offset))
+
     def _get_assigned_names(self, loop_node):
         loop_targets = (ast.For, ast.AsyncFor, ast.comprehension)
         for node in children_in_scope(loop_node):
@@ -805,6 +967,163 @@ class BugBearVisitor(ast.NodeVisitor):
         ):
             self.errors.append(B022(node.lineno, node.col_offset))
 
+    def check_for_b025(self, node):
+        seen = []
+        for handler in node.handlers:
+            if isinstance(handler.type, (ast.Name, ast.Attribute)):
+                name = ".".join(compose_call_path(handler.type))
+                seen.append(name)
+            elif isinstance(handler.type, ast.Tuple):
+                # to avoid checking the same as B014, remove duplicates per except
+                uniques = set()
+                for entry in handler.type.elts:
+                    name = ".".join(compose_call_path(entry))
+                    uniques.add(name)
+                seen.extend(uniques)
+        # sort to have a deterministic output
+        duplicates = sorted({x for x in seen if seen.count(x) > 1})
+        for duplicate in duplicates:
+            self.errors.append(B025(node.lineno, node.col_offset, vars=(duplicate,)))
+
+    def check_for_b905(self, node):
+        if (
+            isinstance(node.func, ast.Name)
+            and node.func.id == "zip"
+            and not any(kw.arg == "strict" for kw in node.keywords)
+        ):
+            self.errors.append(B905(node.lineno, node.col_offset))
+
+    def check_for_b906(self, node: ast.FunctionDef):
+        if not node.name.startswith("visit_"):
+            return
+
+        # extract what's visited, only error if it's a valid ast subclass
+        # with a non-empty _fields attribute - which is what's iterated over in
+        # ast.NodeVisitor.generic_visit
+        class_name = node.name[len("visit_") :]
+        class_type = getattr(ast, class_name, None)
+        if class_type is None or not getattr(class_type, "_fields", None):
+            return
+
+        for n in itertools.chain.from_iterable(ast.walk(nn) for nn in node.body):
+            if isinstance(n, ast.Call) and (
+                (isinstance(n.func, ast.Attribute) and "visit" in n.func.attr)
+                or (isinstance(n.func, ast.Name) and "visit" in n.func.id)
+            ):
+                break
+        else:
+            self.errors.append(B906(node.lineno, node.col_offset))
+
+    def check_for_b028(self, node: ast.JoinedStr):  # noqa: C901
+        # AST structure of strings in f-strings in 3.7 is different enough this
+        # implementation doesn't work
+        if sys.version_info <= (3, 7):
+            return  # pragma: no cover
+
+        def myunparse(node: ast.AST) -> str:  # pragma: no cover
+            if sys.version_info >= (3, 9):
+                return ast.unparse(node)
+            if isinstance(node, ast.Name):
+                return node.id
+            if isinstance(node, ast.Attribute):
+                return myunparse(node.value) + "." + node.attr
+            if isinstance(node, ast.Constant):
+                return repr(node.value)
+            if isinstance(node, ast.Call):
+                return myunparse(node.func) + "()"  # don't bother with arguments
+
+            # as a last resort, just give the type name
+            return type(node).__name__
+
+        quote_marks = "'\""
+        current_mark = None
+        variable = None
+        for value in node.values:
+            # check for quote mark after pre-marked variable
+            if (
+                current_mark is not None
+                and variable is not None
+                and isinstance(value, ast.Constant)
+                and isinstance(value.value, str)
+                and value.value[0] == current_mark
+            ):
+                self.errors.append(
+                    B028(
+                        variable.lineno,
+                        variable.col_offset,
+                        vars=(myunparse(variable.value),),
+                    )
+                )
+                current_mark = variable = None
+                # don't continue with length>1, so we can detect a new pre-mark
+                # in the same string as a post-mark, e.g. `"{foo}" "{bar}"`
+                if len(value.value) == 1:
+                    continue
+
+            # detect pre-mark
+            if (
+                isinstance(value, ast.Constant)
+                and isinstance(value.value, str)
+                and value.value[-1] in quote_marks
+            ):
+                current_mark = value.value[-1]
+                variable = None
+                continue
+
+            # detect variable, if there's a pre-mark
+            if (
+                current_mark is not None
+                and variable is None
+                and isinstance(value, ast.FormattedValue)
+                and value.conversion != ord("r")
+            ):
+                # check if the format spec shows that this is numeric
+                # or otherwise hard/impossible to convert to `!r`
+                if (
+                    isinstance(value.format_spec, ast.JoinedStr)
+                    and value.format_spec.values  # empty format spec is fine
+                ):
+                    if (
+                        # if there's variables in the format_spec, skip
+                        len(value.format_spec.values) > 1
+                        or not isinstance(value.format_spec.values[0], ast.Constant)
+                    ):
+                        current_mark = variable = None
+                        continue
+                    format_specifier = value.format_spec.values[0].value
+
+                    # if second character is an align, first character is a fill
+                    # char - strip it
+                    if len(format_specifier) > 1 and format_specifier[1] in "<>=^":
+                        format_specifier = format_specifier[1:]
+
+                    # strip out precision digits, so the only remaining ones are
+                    # width digits, which will misplace the quotes
+                    format_specifier = re.sub(r"\.\d*", "", format_specifier)
+
+                    # skip if any invalid characters in format spec
+                    invalid_characters = "".join(
+                        (
+                            "=",  # align character only valid for numerics
+                            "+- ",  # sign
+                            "0123456789",  # width digits
+                            "z",  # coerce negative zero floating point to positive
+                            "#",  # alternate form
+                            "_,",  # thousands grouping
+                            "bcdeEfFgGnoxX%",  # various number specifiers
+                        )
+                    )
+                    if set(format_specifier).intersection(invalid_characters):
+                        current_mark = variable = None
+                        continue
+
+                # otherwise save value as variable
+                variable = value
+                continue
+
+            # if no pre-mark or variable detected, reset state
+            current_mark = variable = None
+
 
 def compose_call_path(node):
     if isinstance(node, ast.Attribute):
@@ -826,7 +1145,7 @@ class NameFinder(ast.NodeVisitor):
 
     names = attr.ib(default=attr.Factory(dict))
 
-    def visit_Name(self, node):
+    def visit_Name(self, node):  # noqa: B906 # names don't contain other names
         self.names.setdefault(node.id, []).append(node)
 
     def visit(self, node):
@@ -890,7 +1209,7 @@ class FuntionDefDefaultsVisitor(ast.NodeVisitor):
         # Check for nested functions.
         self.generic_visit(node)
 
-    def visit_Lambda(self, node):
+    def visit_Lambda(self, node):  # noqa: B906
         # Don't recurse into lambda expressions
         # as they are evaluated at call time.
         pass
@@ -931,7 +1250,6 @@ class B020NameFinder(NameFinder):
 error = namedtuple("error", "lineno col message type vars")
 Error = partial(partial, error, type=BugBearChecker, vars=())
 
-
 B001 = Error(
     message=(
         "B001 Do not use {}, it also catches unexpected "
@@ -1081,8 +1399,8 @@ B014.redundant_exceptions = {
 }
 B015 = Error(
     message=(
-        "B015 Pointless comparison. This comparison does nothing but waste "
-        "CPU instructions. Either prepend `assert` or remove it."
+        "B015 Result of comparison is not used. This line doesn't do "
+        "anything. Did you intend to prepend it with assert?"
     )
 )
 B016 = Error(
@@ -1093,16 +1411,17 @@ B016 = Error(
 )
 B017 = Error(
     message=(
-        "B017 assertRaises(Exception): should be considered evil. "
-        "It can lead to your test passing even if the code being tested is "
-        "never executed due to a typo. Either assert for a more specific "
-        "exception (builtin or custom), use assertRaisesRegex, or use the "
-        "context manager form of assertRaises."
+        "B017 assertRaises(Exception): or pytest.raises(Exception) should "
+        "be considered evil. It can lead to your test passing even if the "
+        "code being tested is never executed due to a typo. Either assert "
+        "for a more specific exception (builtin or custom), use "
+        "assertRaisesRegex, or use the context manager form of assertRaises."
     )
 )
 B018 = Error(
     message=(
-        "B018 Found useless expression. Either assign it to a variable or remove it."
+        "B018 Found useless expression. Consider either assigning it to a "
+        "variable or removing it."
     )
 )
 B019 = Error(
@@ -1139,6 +1458,39 @@ B022 = Error(
 )
 
 B023 = Error(message="B023 Function definition does not bind loop variable {!r}.")
+B024 = Error(
+    message=(
+        "B024 {} is an abstract base class, but it has no abstract methods. Remember to"
+        " use the @abstractmethod decorator, potentially in conjunction with"
+        " @classmethod, @property and/or @staticmethod."
+    )
+)
+B025 = Error(
+    message=(
+        "B025 Exception `{0}` has been caught multiple times. Only the first except"
+        " will be considered and all other except catches can be safely removed."
+    )
+)
+B026 = Error(
+    message=(
+        "B026 Star-arg unpacking after a keyword argument is strongly discouraged, "
+        "because it only works when the keyword parameter is declared after all "
+        "parameters supplied by the unpacked sequence, and this change of ordering can "
+        "surprise and mislead readers."
+    )
+)
+B027 = Error(
+    message=(
+        "B027 {} is an empty method in an abstract base class, but has no abstract"
+        " decorator. Consider adding @abstractmethod."
+    )
+)
+B028 = Error(
+    message=(
+        "B028 {!r} is manually surrounded by quotes, consider using the `!r` conversion"
+        " flag."
+    )
+)
 
 # Warnings disabled by default.
 B901 = Error(
@@ -1177,6 +1529,16 @@ B904 = Error(
     )
 )
 
+B905 = Error(message="B905 `zip()` without an explicit `strict=` parameter.")
+
+B906 = Error(
+    message=(
+        "B906 `visit_` function with no further calls to a visit function, which might"
+        " prevent the `ast` visitor from properly visiting all nodes."
+        " Consider adding a call to `self.generic_visit(node)`."
+    )
+)
+
 B950 = Error(message="B950 line too long ({} > {} characters)")
 
-disabled_by_default = ["B901", "B902", "B903", "B904", "B950"]
+disabled_by_default = ["B901", "B902", "B903", "B904", "B905", "B906", "B950"]
