@@ -1,4 +1,5 @@
 import ast
+import itertools
 import os
 import site
 import subprocess
@@ -34,10 +35,17 @@ from bugbear import (
     B021,
     B022,
     B023,
+    B024,
+    B025,
+    B026,
+    B027,
+    B028,
     B901,
     B902,
     B903,
     B904,
+    B905,
+    B906,
     B950,
     BugBearChecker,
     BugBearVisitor,
@@ -246,7 +254,7 @@ class BugbearTestCase(unittest.TestCase):
         filename = Path(__file__).absolute().parent / "b017.py"
         bbc = BugBearChecker(filename=str(filename))
         errors = list(bbc.run())
-        expected = self.errors(B017(22, 8))
+        expected = self.errors(B017(24, 8), B017(26, 8))
         self.assertEqual(errors, expected)
 
     def test_b018_functions(self):
@@ -269,6 +277,14 @@ class BugbearTestCase(unittest.TestCase):
         expected.append(B018(33, 4))
         self.assertEqual(errors, self.errors(*expected))
 
+    def test_b018_modules(self):
+        filename = Path(__file__).absolute().parent / "b018_modules.py"
+        bbc = BugBearChecker(filename=str(filename))
+        errors = list(bbc.run())
+
+        expected = [B018(line, 0) for line in range(9, 19)]
+        self.assertEqual(errors, self.errors(*expected))
+
     def test_b019(self):
         filename = Path(__file__).absolute().parent / "b019.py"
         bbc = BugBearChecker(filename=str(filename))
@@ -293,7 +309,7 @@ class BugbearTestCase(unittest.TestCase):
     def test_b020(self):
         filename = Path(__file__).absolute().parent / "b020.py"
         bbc = BugBearChecker(filename=str(filename))
-        errors = list(bbc.run())
+        errors = [e for e in bbc.run() if e[2][:4] == "B020"]
         self.assertEqual(
             errors,
             self.errors(
@@ -347,9 +363,174 @@ class BugbearTestCase(unittest.TestCase):
             B023(61, 16, vars=("j",)),
             B023(61, 20, vars=("k",)),
             B023(68, 9, vars=("l",)),
+            B023(113, 23, vars=("x",)),
+            B023(114, 26, vars=("x",)),
+            B023(115, 36, vars=("x",)),
+            B023(116, 37, vars=("x",)),
+            B023(117, 36, vars=("x",)),
+            B023(168, 28, vars=("name",)),  # known false alarm
+            B023(171, 28, vars=("i",)),
         )
         self.assertEqual(errors, expected)
 
+    def test_b024(self):
+        filename = Path(__file__).absolute().parent / "b024.py"
+        bbc = BugBearChecker(filename=str(filename))
+        errors = list(bbc.run())
+        expected = self.errors(
+            B024(17, 0, vars=("Base_1",)),
+            B024(52, 0, vars=("Base_7",)),
+            B024(58, 0, vars=("MetaBase_1",)),
+            B024(69, 0, vars=("abc_Base_1",)),
+            B024(74, 0, vars=("abc_Base_2",)),
+            B024(128, 0, vars=("abc_set_class_variable_4",)),
+        )
+        self.assertEqual(errors, expected)
+
+    def test_b025(self):
+        filename = Path(__file__).absolute().parent / "b025.py"
+        bbc = BugBearChecker(filename=str(filename))
+        errors = list(bbc.run())
+        self.assertEqual(
+            errors,
+            self.errors(
+                B025(15, 0, vars=("ValueError",)),
+                B025(22, 0, vars=("pickle.PickleError",)),
+                B025(31, 0, vars=("TypeError",)),
+                B025(31, 0, vars=("ValueError",)),
+            ),
+        )
+
+    def test_b026(self):
+        filename = Path(__file__).absolute().parent / "b026.py"
+        bbc = BugBearChecker(filename=str(filename))
+        errors = list(bbc.run())
+        self.assertEqual(
+            errors,
+            self.errors(
+                B026(16, 15),
+                B026(17, 15),
+                B026(18, 26),
+                B026(19, 37),
+                B026(20, 15),
+                B026(20, 25),
+                B026(21, 25),
+            ),
+        )
+
+    def test_b027(self):
+        filename = Path(__file__).absolute().parent / "b027.py"
+        bbc = BugBearChecker(filename=str(filename))
+        errors = list(bbc.run())
+        expected = self.errors(
+            B027(13, 4, vars=("empty_1",)),
+            B027(16, 4, vars=("empty_2",)),
+            B027(19, 4, vars=("empty_3",)),
+            B027(23, 4, vars=("empty_4",)),
+            B027(31 if sys.version_info >= (3, 8) else 30, 4, vars=("empty_5",)),
+        )
+        self.assertEqual(errors, expected)
+
+    @unittest.skipIf(sys.version_info < (3, 8), "not implemented for <3.8")
+    def test_b028(self):
+        filename = Path(__file__).absolute().parent / "b028.py"
+        bbc = BugBearChecker(filename=str(filename))
+        errors = list(bbc.run())
+        expected = self.errors(
+            B028(8, 0, vars=("var",)),
+            B028(9, 0, vars=("var",)),
+            B028(10, 0, vars=("var",)),
+            B028(12, 0, vars=("var",)),
+            B028(13, 0, vars=("var",)),
+            B028(14, 0, vars=("var",)),
+            B028(16, 0, vars=("'hello'",)),
+            B028(17, 0, vars=("foo()",)),
+            B028(20, 5, vars=("var",)),
+            B028(25, 5, vars=("var",)),
+            B028(31, 0, vars=("var",)),
+            B028(32, 0, vars=("var",)),
+            B028(33, 0, vars=("var",)),
+            B028(33, 0, vars=("var2",)),
+            B028(34, 0, vars=("var",)),
+            B028(34, 0, vars=("var2",)),
+            B028(35, 0, vars=("var",)),
+            B028(35, 0, vars=("var2",)),
+            B028(38, 0, vars=("var2",)),
+            B028(41, 0, vars=("var",)),
+            B028(42, 0, vars=("var.__str__",)),
+            B028(43, 0, vars=("var.__str__.__repr__",)),
+            B028(44, 0, vars=("3 + 5" if sys.version_info >= (3, 9) else "BinOp",)),
+            B028(45, 0, vars=("foo()",)),
+            B028(46, 0, vars=("None",)),
+            B028(47, 0, vars=("..." if sys.version_info >= (3, 9) else "Ellipsis",)),
+            B028(48, 0, vars=("True",)),
+            B028(51, 0, vars=("var",)),
+            B028(52, 0, vars=("var",)),
+            B028(53, 0, vars=("var",)),
+            B028(54, 0, vars=("var",)),
+            B028(57, 0, vars=("var",)),
+            B028(60, 0, vars=("var",)),
+            B028(64, 0, vars=("var",)),
+            B028(66, 0, vars=("var",)),
+            B028(68, 0, vars=("var",)),
+        )
+        self.assertEqual(errors, expected)
+
+    # manual permutations to save overhead when doing >60k permutations
+    # see format spec at
+    # https://docs.python.org/3/library/string.html#format-specification-mini-language
+    @unittest.skipIf(sys.version_info < (3, 8), "not implemented for <3.8")
+    def test_b028_format_specifier_permutations(self):
+        visitor = BugBearVisitor(filename="", lines="")
+
+        for fields in itertools.product(
+            (None, "x"),  # fill (any character)
+            (None, *"<>=^"),  # align
+            (None, *"+- "),  # sign
+            (None, "z"),  # z_letter
+            (None, "#"),  # pound_sign
+            (None, "0"),  # zero
+            (None, *"19"),  # width
+            (None, *"_,"),  # grouping_option
+            (None, ".8"),  # precision
+            (None, *"bcdeEfFgGnosxX%"),  # type_
+        ):
+            format_spec = "".join(f for f in fields if f is not None)
+
+            # directly interact with a visitor to save on runtime
+            bbc_string = "f'\"{var:" + format_spec + "}\"'"
+            tree = ast.parse(bbc_string)
+            visitor.errors = []
+            visitor.visit(tree)
+
+            format_string = "'{:" + format_spec + "}'"
+            try:
+                old = format_string.format("hello")
+            except ValueError:
+                assert (
+                    visitor.errors == []
+                ), f"b028 raised for {format_spec!r} not valid for string"
+                continue
+
+            new = ("{!r:" + format_spec + "}").format("hello")
+
+            # Preceding the width field by 0 in >=3.10 is valid, but does nothing.
+            # The presence of it means likely numeric variable though.
+            # A width shorter than the string will look the same, but should not give b028.
+            if fields[5] == "0" or fields[6] == "1":
+                assert (
+                    visitor.errors == []
+                ), f"b028 should not raise on questionable case {format_spec}"
+            elif old == new:
+                assert visitor.errors, (
+                    f"b028 not raised for {format_spec} that would look identical"
+                    " with !r"
+                )
+            else:
+                assert (
+                    visitor.errors == []
+                ), f"b028 raised for {format_spec} that would look different with !r"
+
     def test_b901(self):
         filename = Path(__file__).absolute().parent / "b901.py"
         bbc = BugBearChecker(filename=str(filename))
@@ -415,6 +596,31 @@ class BugbearTestCase(unittest.TestCase):
         ]
         self.assertEqual(errors, self.errors(*expected))
 
+    @unittest.skipIf(sys.version_info < (3, 10), "requires 3.10+")
+    def test_b905(self):
+        filename = Path(__file__).absolute().parent / "b905_py310.py"
+        bbc = BugBearChecker(filename=str(filename))
+        errors = list(bbc.run())
+        expected = [
+            B905(1, 0),
+            B905(2, 0),
+            B905(3, 0),
+            B905(4, 0),
+            B905(4, 15),
+            B905(5, 4),
+            B905(6, 0),
+        ]
+        self.assertEqual(errors, self.errors(*expected))
+
+    def test_b906(self):
+        filename = Path(__file__).absolute().parent / "b906.py"
+        bbc = BugBearChecker(filename=str(filename))
+        errors = list(bbc.run())
+        expected = [
+            B906(9, 0),
+        ]
+        self.assertEqual(errors, self.errors(*expected))
+
     def test_b950(self):
         filename = Path(__file__).absolute().parent / "b950.py"
         bbc = BugBearChecker(filename=str(filename))
